﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;

namespace InstanceFactory.CustomPrivateAccessor.ConsoleApplication.Test.TestTools.UnitTesting
{
  /// <summary>
  /// Base class for private accessors that provide access to non-public members of classes.
  /// </summary>
  /// <typeparam name="T">Type of class to provide access to.</typeparam>
  internal abstract class PrivateAccessorBase<T>
  {
    #region Private Const Data Member

    /// <summary>
    /// Keeps the prefix of property getter method names.
    /// </summary>
    private const string GetterPrefix = "get_";

    /// <summary>
    /// Keeps the prefix of property setter method names.
    /// </summary>
    private const string SetterPrefix = "set_";

    #endregion Private Const Data Member


    #region Private Properties

    /// <summary>
    /// Gets / sets the accessor to the object to access non-public member.
    /// </summary>
    private PrivateObject PrivateObject { get; set; }

    #endregion Private Properties


    #region Private Static Properties

    /// <summary>
    /// Gets / sets the declared proporties for later use, in case static properties should be accessed.
    /// </summary>
    private static IEnumerable<PropertyInfo> DeclaredProperties { get; set; }

    #endregion Private Static Properties


    #region Public Properties

    /// <summary>
    /// Gets the instance of the class to be used for accessing non-public members.
    /// </summary>
    public T TestInstance { get { return ((T)PrivateObject.Target); } }

    #endregion Public Properties


    #region Static Constructor

    /// <summary>
    /// Static initialization.
    /// </summary>
    static PrivateAccessorBase()
    {
      // Get the declared properties for later use, in case static properties should be accessed.
      PrivateAccessorBase<T>.DeclaredProperties = typeof(T).GetProperties(BindingFlags.NonPublic | BindingFlags.Static);
    }

    #endregion Static Constructor


    #region Internal Constructors

    /// <summary>
    /// Initializes a new instance of <see cref="PrivateAccessorBase"/> with an instance of the class to be tested.
    /// </summary>
    /// <param name="instance">Instance of the class to be tested.</param>
    internal PrivateAccessorBase
      (
      T instance
      )
    {
      if (instance == null)
      {
        throw (new ArgumentNullException("instance", "The instance to test must not be null."));
      }

      // Create the PrivateObject.
      PrivateObject = new PrivateObject(instance);
    }

    #endregion Internal Constructors


    #region Protected Methods

    /// <summary>
    /// Returns the value of a property. The name of the property is taken from the call stack (for .NET version < 4.5).
    /// </summary>
    /// <typeparam name="T">Type of the return value.</typeparam>
    /// <returns>The property value.</returns>
    [MethodImpl(MethodImplOptions.NoInlining)]
    protected TReturnValue GetPropertyValue<TReturnValue>()
    {
      // Need to remove the "get_" prefix from the caller's name
      string propertyName = new StackFrame(1, true).GetMethod().Name.Replace(PrivateAccessorBase<T>.GetterPrefix, String.Empty);

      return ((TReturnValue)PrivateObject.GetProperty(propertyName, null));
    }

    /// <summary>
    /// Sets the value of a property. The name of the property is taken from the call stack (for .NET version < 4.5).
    /// </summary>
    /// <param name="value">The value to be set.</param>
    [MethodImpl(MethodImplOptions.NoInlining)]
    protected void SetPropertyValue
      (
      object value
      )
    {
      // Need to remove the "set_" prefix from the caller's name
      string propertyName = new StackFrame(1, true).GetMethod().Name.Replace(PrivateAccessorBase<T>.SetterPrefix, String.Empty);

      PrivateObject.SetProperty(propertyName, value, new object[] { });
    }

    /// <summary>
    /// Invokes a method with parameter an return value.
    /// </summary>
    /// <typeparam name="TReturnValue">The type of the return value.</typeparam>
    /// <param name="parameter">The parameter to be passed.</param>
    [MethodImpl(MethodImplOptions.NoInlining)]
    protected TReturnValue Invoke<TReturnValue>
      (
      params object[] parameter
      )
    {
      // Take the caller's name as method name
      return ((TReturnValue)PrivateObject.Invoke(new StackFrame(1, true).GetMethod().Name, parameter));
    }

    #endregion Protected Methods


    #region Protected Static Methods

    /// <summary>
    /// Returns the value of a static property. The name of the property is taken from the call stack (for .NET version < 4.5).
    /// </summary>
    /// <typeparam name="T">Type of the return value.</typeparam>
    /// <returns>The property value.</returns>
    [MethodImpl(MethodImplOptions.NoInlining)]
    protected static TReturnValue GetStaticPropertyValue<TReturnValue>()
    {
      // Need to remove the "get_" prefix from the caller's name
      string propertyName = new StackFrame(1, true).GetMethod().Name.Replace(PrivateAccessorBase<T>.GetterPrefix, String.Empty);

      // Find the property having the matching name and get the value
      return ((TReturnValue)PrivateAccessorBase<T>.DeclaredProperties.Single(info => info.Name.Equals(propertyName)).GetValue(null, null));
    }

    /// <summary>
    /// Sets the value of a static property. The name of the property is taken from the call stack (for .NET version < 4.5).
    /// </summary>
    /// <param name="value">The value to be set.</param>
    /// <returns>The property value.</returns>
    [MethodImpl(MethodImplOptions.NoInlining)]
    protected static void SetStaticPropertyValue
      (
      object value
      )
    {
      // Need to remove the "set_" prefix from the caller's name
      string propertyName = new StackFrame(1, true).GetMethod().Name.Replace(PrivateAccessorBase<T>.SetterPrefix, String.Empty);

      // Find the property having the matching name and set the value
      PrivateAccessorBase<T>.DeclaredProperties.Single(info => info.Name.Equals(propertyName)).SetValue(null, value, null);
    }

    #endregion Protected Static Methods
  }
}
